package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsFactory;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsWriter;
import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupLogger;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.ui.cli2.instruction.*;
import be.iminds.ilabt.jfed.ui.cli2.logging.ActionLogger;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.encoder.PatternLayoutEncoder;
import ch.qos.logback.classic.filter.ThresholdFilter;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.FileAppender;
import ch.qos.logback.core.OutputStreamAppender;
import ch.qos.logback.core.util.StatusPrinter;
import com.google.inject.Injector;
import javanet.staxutils.IndentingXMLEventWriter;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Context: All derived and variable state related to executing the Instruction
 *
 * This contains getters/setters for:
 *   - injector and related things
 *   - project info
 *   - slice info
 *   - rspec info
 *   - sliver info (if relevant)
 *
 * This contains methods for:
 *   - logger setup and access
 *   - call logging setup and access
 *   - action result output setup and access
 *
 */
public class Context {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(Context.class);

    @Nonnull public final String[] args;

    @Nonnull private final Instruction instruction;
    @Nonnull private final ExperimenterCommandLineParser.GenericCliArguments genericCliArguments;

    //System.{in,out,err}, or a stream that is pretending to be it
    @Nonnull private final PrintStream out;
    @Nonnull private final PrintStream err;
    @Nonnull private final InputStream in;

    @Nonnull private final BaseCli baseCli;

    private final List<Closeable> toClose = new ArrayList<>();

    private Injector injector;

    private String sliceUrn;
    private String sliceName;
    private String projectName;

    @Nonnull private final ActionLogger actionLogger;

    private ExitReason exitReason = null;

    public Context(@Nonnull String[] args,
                   @Nonnull ExperimenterCommandLineParser.GenericCliArguments genericCliArguments,
                   @Nonnull Instruction instruction,
                   @Nonnull PrintStream out,
                   @Nonnull PrintStream err,
                   @Nonnull InputStream in,
                   @Nonnull BaseCli baseCli) {
        this.args = args;
        this.instruction = instruction;
        this.genericCliArguments = genericCliArguments;
        this.out = out;
        this.err = err;
        this.in = in;
        this.baseCli = baseCli;
        this.actionLogger = new ActionLogger();
    }

    @Nonnull
    public Instruction getInstruction() {
        return instruction;
    }

    @Nonnull
    public ExperimenterCommandLineParser.GenericCliArguments getGenericCliArguments() {
        return genericCliArguments;
    }

    @Nonnull
    public ActionLogger getActionLogger() {
        return actionLogger;
    }

    @Nonnull
    public ExperimentSetupLogger getExperimentSetupLogger() {
        org.slf4j.Logger debugLogger = LOG;
        return new ExperimentSetupLogger() {

            @Override
            public void fatal(@Nullable String val, @Nullable Throwable t) {
                debugLogger.error(val, t);
                exitWithCode(ExitReason.ERROR, val);
            }

            @Override
            public void logExtraText(@Nonnull String name, @Nonnull Object value, boolean large) {
                debugLogger.info(name+": "+value);
            }

            @Override
            public void logExtraXml(@Nonnull String name, @Nonnull Object value, boolean large) {
                debugLogger.info(name+": "+value);
            }

            @Override
            public void logExtraHtml(@Nonnull String name, @Nonnull Object value, boolean large) {
                debugLogger.info(name+": "+value);
            }

            @Override
            public void logExtraJson(@Nonnull String name, @Nonnull Object value, boolean large) {
                debugLogger.info(name+": "+value);
            }

            @Override
            public void debug(@Nullable String val, @Nullable Throwable e, boolean formatAsCode) {
                debugLogger.debug(val, e);
            }

            @Override
            public void info(@Nullable String val, @Nullable Throwable e, boolean formatAsCode) {
                debugLogger.info(val, e);
            }

            @Override
            public void warn(@Nullable String val, @Nullable Throwable e, boolean formatAsCode) {
                debugLogger.warn(val, e);
            }

            @Override
            public void error(@Nullable String val, @Nullable Throwable e, boolean formatAsCode) {
                debugLogger.error(val, e);
            }
        };
    }


    //not exposed: use instead:
    //      - for debug output: just use slf4j logging
    //      - for call output: getLogger()
    //      - for action output: actionPrint / actionPrintln
//    @Nonnull
//    public PrintStream getOut() {
//        return out;
//    }
//
//    @Nonnull
//    public PrintStream getErr() {
//        return err;
//    }
//
//    @Nonnull
//    public InputStream getIn() {
//        return in;
//    }

    public enum ExitReason {
        SUCCESS(0),
        ERROR(1),                 //any other error
        NOT_ENOUGH_RESOURCES(2),
        BAD_ARGS(3),
        SERVER_ERROR(4),          //any other server side error
        INTERRUPTED(10);          //interrupted by user somehow (CTRL-C or other)

        private final int exitVal;

        ExitReason(int exitVal) {
            this.exitVal = exitVal;
        }

        public int getExitVal() {
            return exitVal;
        }
    }
    //thread safe. Triggers next checkExit to throw
    public void exitWithCode(ExitReason code, String message) {
        this.exitReason = code;
        actionLogger.errorPrintln(message);
    }
    //throws ExitException if exitWithCode was called
    public void throwIfExited() throws CliExitException {
        if (this.exitReason != null)
            throw new CliExitException(exitReason);
    }


    //cli.getErr();
    //cli.getOut();
    //cli.getIn();

    public void closeOutputs() {
        for (Closeable close : toClose) {
            try {
                close.close();
            } catch (IOException e) {
                LOG.warn("Error while closing", e);
            }
        }
    }

    public void setupActionOutput() throws CliExitException {
        if (instruction.getActionOutputs().isEmpty()) {
            actionLogger.addESpecLogListener(new ActionLogger.StreamActionLogListener(out, err, true, false));
            return;
        }

        for (ActionOutput actionOutput : instruction.getActionOutputs()) {
            boolean doNonJsonOut = actionOutput.getFormat() == ActionOutput.Format.TXT;
            boolean doJsonOut = actionOutput.getFormat() == ActionOutput.Format.JSON;

            switch (actionOutput.getTarget()) {
                case STDERR: {
                    actionLogger.addESpecLogListener(new ActionLogger.StreamActionLogListener(err, err, doNonJsonOut, doJsonOut));
                    break;
                }
                case STDOUT: {
                    actionLogger.addESpecLogListener(new ActionLogger.StreamActionLogListener(out, err, doNonJsonOut, doJsonOut));
                    break;
                }
                case FILE: {
                    if (actionOutput.getFilename() == null)
                        throw new CliExitException(ExitReason.BAD_ARGS, "Could action output type FILE requires a filename");
                    try {
                        PrintStream fout = new PrintStream(new FileOutputStream(actionOutput.getFilename()));
                        toClose.add(fout);
                        actionLogger.addESpecLogListener(new ActionLogger.StreamActionLogListener(fout, err, doNonJsonOut, doJsonOut));
                    } catch (FileNotFoundException e) {
                        throw new CliExitException(ExitReason.BAD_ARGS, "Could not open action output file \""+actionOutput.getFilename()+"\"");
                    }
                    break;
                }
                default:
                    throw new IllegalStateException("Should not be called for " + actionOutput.getTarget());
            }
        }
    }

    /**
     * Usefull for debugging things before debug output is setup
     */
    public static void setupDebugOutputToStdout() {
        LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
        assert loggerContext != null;
        loggerContext.reset();

        ch.qos.logback.classic.Logger rootLogger = loggerContext.getLogger(org.slf4j.Logger.ROOT_LOGGER_NAME);

        PatternLayoutEncoder encoder = new PatternLayoutEncoder();
        encoder.setContext(loggerContext);
        encoder.setPattern("%-4relative %-5level %logger{30} - %msg%n");
        encoder.start();

        final OutputStreamAppender<ILoggingEvent> appender;
        OutputStreamAppender<ILoggingEvent> osAppender = new OutputStreamAppender<>();
        osAppender.setContext(loggerContext);
        osAppender.setEncoder(encoder);
        osAppender.setName("stdout appender");
        osAppender.setOutputStream(System.out);
        appender = osAppender;

        assert appender.getEncoder() != null;
        assert appender.getContext() != null;

        ThresholdFilter thresholdFilter = new ThresholdFilter();
        thresholdFilter.setContext(loggerContext);
        thresholdFilter.setLevel("DEBUG");
        thresholdFilter.start();

        appender.addFilter(thresholdFilter);
        appender.start();

        rootLogger.addAppender(appender);

        loggerContext.getLogger("org.apache.http").setLevel(Level.WARN);
        loggerContext.getLogger("be.iminds.ilabt.jfed.highlevel.controller").setLevel(Level.INFO);
        loggerContext.getLogger("be.iminds.ilabt.jfed.lowlevel.connection").setLevel(Level.ERROR);

        rootLogger.setLevel(Level.DEBUG);

        StatusPrinter.printInCaseOfErrorsOrWarnings(loggerContext);
    }

    public void setupDebugOutput() {
        if (instruction.getDebugOutputs().isEmpty()) {
            //logging is a bit nasty, as it is static.
            //This is only a problem if there is another java program that wants to start the CLI
            //directly (not a seperate or sub process). In that case, either:
            //   - We change the logging, but also change it for the "parent" program
            //   - We don't change the logging, and it is the same (and mixed with) the logging of the parent program
            //
            // (you want neither, but you'll have to choose one.)
            //
            // This is a hard problem, without decent solution.
            // See also: https://logback.qos.ch/manual/loggingSeparation.html

            LOG.info("setupDebugOutput() when no debug output is configured. \n" +
                    "This will fall back to NOT re-configuring logback-classic. \n" +
                    "In case of calling using CLI directly, this defaults to the logback.xml file, " +
                    "which does no logging. \n" +
                    "In case the CLI code was called from another java program, the logging of that program " +
                    "will be used.");
            return;
        }

        LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
        assert loggerContext != null;
//        LoggerContext loggerContext = new LoggerContext(); //alternative that does not change global config.

//        JoranConfigurator configurator = new JoranConfigurator();
//        configurator.setContext(loggerContext);
        loggerContext.reset();

        ch.qos.logback.classic.Logger rootLogger = loggerContext.getLogger(org.slf4j.Logger.ROOT_LOGGER_NAME);

        DebugOutput.Level minLevel = DebugOutput.Level.ERROR;
        for (DebugOutput debugOutput : instruction.getDebugOutputs()) {
            if (debugOutput.getLevel().getRank() < minLevel.getRank())
                minLevel = debugOutput.getLevel();

            PatternLayoutEncoder encoder = new PatternLayoutEncoder();
            encoder.setContext(loggerContext);
            switch (debugOutput.getFormat()) {
                case TXT: {
                    encoder.setPattern("%-4relative %-5level %logger{30} - %msg%n");
                    break;
                }
                case TXT_FULL: {
                    encoder.setPattern("%d{HH:mm:ss.SSS} [%thread] %-5level %logger{40} - %msg%n");
                    break;
                }
            }
//            encoder.setPattern("%d{HH:mm:ss.SSS} [%thread] %-5level %logger{36} - %msg%n");
            encoder.start();

            final OutputStreamAppender<ILoggingEvent> appender;
            if (debugOutput.getTarget() == OutputTarget.STDOUT || debugOutput.getTarget() == OutputTarget.STDERR) {
                OutputStreamAppender<ILoggingEvent> osAppender = new OutputStreamAppender<>();
                osAppender.setContext(loggerContext);
                osAppender.setEncoder(encoder);
                switch (debugOutput.getTarget()) {
                    case STDERR: {
                        osAppender.setName("stderr "+debugOutput.getFormat()+" appender");
                        osAppender.setOutputStream(err);
                        break;
                    }
                    case STDOUT: {
                        osAppender.setName("stdout "+debugOutput.getFormat()+" appender");
                        osAppender.setOutputStream(out);
                        break;
                    }
                    default:
                        throw new IllegalStateException("Should not be called for " + debugOutput.getTarget());
                }
                appender = osAppender;
            } else {
                assert debugOutput.getTarget() == OutputTarget.FILE;
                FileAppender<ILoggingEvent> fileAppender = new FileAppender<>();
                fileAppender.setContext(loggerContext);
                fileAppender.setEncoder(encoder);
                if (debugOutput.getFilename() == null)
                    throw new IllegalStateException("if target debug output is FILE, you must specify a filename");
                fileAppender.setFile(debugOutput.getFilename());
                appender = fileAppender;
            }

            assert appender.getEncoder() != null;
            assert appender.getContext() != null;

            ThresholdFilter thresholdFilter = new ThresholdFilter();
            thresholdFilter.setContext(loggerContext);
            thresholdFilter.setLevel(debugOutput.getLevel().getLogbackLevel().toString());
            thresholdFilter.start();

            appender.addFilter(thresholdFilter);
            appender.start();

            rootLogger.addAppender(appender);
        }

        //  <logger name="org.apache.http" level="warn"/>
        //  <logger name="be.iminds.ilabt.jfed.highlevel.controller" level="info"/>
        //  <logger name="be.iminds.ilabt.jfed.lowlevel.connection" level="error"/>
        loggerContext.getLogger("org.apache.http").setLevel(Level.WARN);
        loggerContext.getLogger("be.iminds.ilabt.jfed.highlevel.controller").setLevel(Level.INFO);
        loggerContext.getLogger("be.iminds.ilabt.jfed.lowlevel.connection").setLevel(Level.ERROR);

        rootLogger.setLevel(minLevel.getLogbackLevel());

        StatusPrinter.printInCaseOfErrorsOrWarnings(loggerContext);
    }

    public void setupInjector() throws CliExitException {
        baseCli.initAuthoritiesModule(null);
        try {
            //TODO: password handling is only partially implemented,
            //      as there are different methods to get the password.
            //      We should probably pass a supplier around instead.
            assert instruction.getUser().getPasswordMethod() != User.PasswordMethod.FILE;
            assert instruction.getUser().getPasswordMethod() != User.PasswordMethod.STDIN;

            baseCli.initUserModule(instruction.getUser().getPem(),
                    genericCliArguments.getContextFilename(),
                    instruction.getUser().getPassword(),
                    false,
                    in);
        } catch (IOException e) {
            LOG.debug("Error initialising user module", e);
            throw new CliExitException(ExitReason.ERROR, "Error initialising user module", e);
        } catch (BaseCli.CliArgumentException e) {
            LOG.debug("CliArgumentException initialising user module", e);
            throw new ArgumentException(e.getMessage(), e);
        }

        injector = baseCli.getInjector(actionLogger::errorPrintln);
        GeniUser user = injector.getInstance(GeniUser.class);

        if (!BaseCli.checkUser(user, actionLogger::errorPrintln)) {
            exitWithCode(Context.ExitReason.BAD_ARGS, "Problem with user certificate.");
            throwIfExited();
        }
    }

    public void setupProxy() {
        assert instruction.getProxy() != null;
        if (instruction.getProxy().getUseForCalls() == null && instruction.getProxy().getUseForSsh() == null) {
            return;
        }

        ProxyPreferencesManager proxyPreferencesManager = injector.getInstance(ProxyPreferencesManager.class);
        JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);

        if (instruction.getProxy().getUseForCalls() != null) {
            if (instruction.getProxy().getUseForCalls())
                jFedPreferences.setString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED, ProxyPreferencesManager.ALWAYS_USE_PROXY);
            else
                jFedPreferences.setString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED, ProxyPreferencesManager.NEVER_USE_PROXY);
            proxyPreferencesManager.updateProxySettings();
        }
        if (instruction.getProxy().getUseForSsh() != null) {
            if (instruction.getProxy().getUseForSsh())
                jFedPreferences.setString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH, ProxyPreferencesManager.ALWAYS_USE_PROXY);
            else
                jFedPreferences.setString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH, ProxyPreferencesManager.NEVER_USE_PROXY);
            //no need to updateProxySettings()
        }
    }

    public Injector getInjector() {
        return injector;
    }

    @Nonnull
    public GeniUser getUser() {
        GeniUser user = injector.getInstance(GeniUser.class);
        if (user == null) {
            exitWithCode(Context.ExitReason.ERROR, "Problem with GeniUser.");
            throw new RuntimeException("GeniUser should never be null at this point");
        }
        return user;
    }
    @Nonnull
    public GeniUserProvider getUserProvider() {
        GeniUserProvider res = injector.getInstance(GeniUserProvider.class);
        if (res == null) {
            exitWithCode(Context.ExitReason.ERROR, "Problem with GeniUserProvider.");
            throw new RuntimeException("GeniUserProvider should never be null at this point");
        }
        return res;
    }

    public void setupCallOutput() throws CliExitException {
        Logger logger = injector.getInstance(Logger.class);

        for (CallOutput callOutput : instruction.getCallOutputs()) {
            CallOutputHandler callOutputHandler;

            switch (callOutput.getFormat()) {
                case XML: {
                    callOutputHandler = new CallOutputXmlHandler(callOutput);
                    break;
                }
                case TEXT: {
                    callOutputHandler = new CallOutputPlainTextHandler(callOutput);
                    break;
                }
//                case JSON: { //TODO implement CallOutputJsonHandler (see CallOutputPlainTextHandler)
//                    callOutputHandler = new CallOutputJsonHandler(callOutput);
//                    break;
//                }
//                case HTML: { //TODO implement CallOutputHtmlHandler
//                    callOutputHandler = new CallOutputHtmlHandler(callOutput);
//                    break;
//                }
                default:
                    throw new RuntimeException("call log output \""+callOutput.getFormat()+"\" is not supported");
            }

            toClose.add(callOutputHandler);
            logger.addResultListener(callOutputHandler);
        }
    }


    @Nonnull
    public BaseCli getBaseCli() {
        return baseCli;
    }

    public String getSliceUrn() {
        return sliceUrn;
    }

    public void setSliceUrn(String sliceUrn) {
        this.sliceUrn = sliceUrn;
    }

    public String getSliceName() {
        return sliceName;
    }

    public void setSliceName(String sliceName) {
        this.sliceName = sliceName;
    }

    public String getProjectName() {
        return projectName;
    }

    public void setProjectName(String projectName) {
        this.projectName = projectName;
    }


    private abstract class CallOutputHandler implements ResultListener, Closeable {
        @Nonnull protected final OutputStream outputStream;

        public CallOutputHandler(CallOutput callOutput) throws CliExitException {
            switch (callOutput.getTarget()) {
                case FILE: {
                    if (callOutput.getFilename() == null)
                        throw new IllegalArgumentException("callOutput.getFilename() may not be null when " +
                                "callOutput.getTarget()="+callOutput.getTarget());
                    try {
                        this.outputStream = new FileOutputStream(callOutput.getFilename());
                    } catch (FileNotFoundException e) {
                        LOG.error("Could not open call output file", e);
                        throw new ArgumentException("Could not open call output file " +
                                "\""+callOutput.getFilename()+"\" for writing: "+e.getMessage());
                    }
                    break;
                }
                case STDOUT: {
                    outputStream = out;
                    break;
                }
                case STDERR: {
                    outputStream = err;
                    break;
                }
                default:
                    throw new IllegalArgumentException("callOutput.target="+callOutput.getTarget()+" " +
                            "is not supported");
            }
        }

        public abstract void flush();
    }

    private class CallOutputXmlHandler extends CallOutputHandler {
        protected final XMLEventWriter xtw;
        protected final XMLEventFactory eventFactory;

        private CallOutputXmlHandler(CallOutput callOutput) throws CliExitException {
            super(callOutput);
            try {

                XMLOutputFactory xof = XMLOutputFactory.newInstance();
                xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(outputStream, "UTF-8"));
                eventFactory = XMLEventFactory.newInstance();

                xtw.add(eventFactory.createStartDocument());
                xtw.add(eventFactory.createStartElement(new QName("calls"), null, null));
            } catch (XMLStreamException e) {
                LOG.error("Exception generating XML for call output", e);
                close();
                throw new ArgumentException("Failed to initialise call output " +
                        callOutput+"", e);
            }
        }

        @Override
        public void flush() {
            try {
                xtw.flush();
                outputStream.flush();
            } catch (Exception e) {
                LOG.warn("Unexpected error flusing", e);
            }
        }

        @Override
        public void close() {
            if (xtw != null && eventFactory != null) {
                try {
                    xtw.add(eventFactory.createEndElement(new QName("calls"), null));
                    xtw.add(eventFactory.createEndDocument());
                    xtw.flush();
                } catch (XMLStreamException e) {
                    LOG.error("Exception while closing call output", e);
                }
            }
            //noinspection ObjectEquality
            if (outputStream != out && outputStream != err) {
                try {
                    outputStream.close();
                } catch (IOException e) {
                    LOG.error("Exception while closing call output file", e);
                }
            }
        }

        @Override
        public void onResult(ApiCallDetails result) {
            SerializableApiCallDetails serializableApiCallDetails =
                    new SerializableApiCallDetailsFactory()
                            .createFromApiCallDetails(result);

            try {
                if (result.getServerUrnString() != null)
                    xtw.add(eventFactory.createComment("" + result.getGeniMethodName() + " on " +
                            result.getServerUrnString() + ""));

                SerializableApiCallDetailsWriter.writeApiCallDetails(serializableApiCallDetails, xtw, eventFactory);

                //add 2 newlines
                xtw.add(eventFactory.createCharacters(System.getProperty("line.separator") +
                        System.getProperty("line.separator")));

                xtw.flush();
                outputStream.flush();
            } catch (IOException e) {
                LOG.error("Error flushing", e);
            } catch (XMLStreamException e) {
                LOG.error("Error writing call to stream", e);
                close();
            }
        }
    }

    private class CallOutputPlainTextHandler extends CallOutputHandler {
        @Nonnull private final PrintStream printCallsOut;
        private final static String printCallsPrefix = "CALL: ";

        public CallOutputPlainTextHandler(CallOutput callOutput) throws CliExitException {
            super(callOutput);
            this.printCallsOut = new PrintStream(outputStream);
        }

        @Override
        public void flush() {
            try {
                outputStream.flush();
            } catch (IOException e) {
                LOG.warn("Error flushing output stream", e);
            }
        }

        @Override
        public void close() {
            //noinspection ObjectEquality
            if (outputStream != out && outputStream != err) {
                try {
                    outputStream.close();
                } catch (IOException e) {
                    LOG.error("Exception while closing call output file", e);
                }
            }
        }

        @Override
        public void onResult(ApiCallDetails result) {
            printCallsOut.println();
            printCallsOut.println(TextUtil.indent(printCallsPrefix.length()+1, printCallsPrefix + "" + result.getGeniMethodName() +
                    " on " + result.getServerUrnString() + "" +
                    " at "+result.getConnectionConfig().getServerUrlString()));

            String req = result.getXmlRpcRequestJsonString();
            String res = result.getXmlRpcResponseJsonString();
            if (req != null) printCallsOut.println(printCallsPrefix + "Request: " + req);
            if (res != null) printCallsOut.println(printCallsPrefix + "Reply: " + res);

            printCallsOut.println();
            printCallsOut.println();
            printCallsOut.flush();
        }
    }


}
